%% tclp__utils.pl: misc. predicates used in tclp
%% Copyright (C) 2001,2002 INRIA Emmanuel Coquery

%% This library is free software; you can redistribute it and/or
%% modify it under the terms of the GNU Lesser General Public
%% License as published by the Free Software Foundation; either
%% version 2.1 of the License, or (at your option) any later version.

%% This library is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
%% Lesser General Public License for more details.

%% You should have received a copy of the GNU Lesser General Public
%% License along with this library; if not, write to the Free Software
%% Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%                                                                           %%
%%                                   TCLP                                    %%
%%                                                                           %%
%%         Emmanuel Coquery, projet Contraintes, INRIA Rocquencourt          %%
%%                                                                           %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% $Header: /home/beaune/contraintes/coquery/cvs/contraintes/tclpchr/src/tclp/tclp__utils.pl,v 1.10 2004/01/23 14:04:33 coquery Exp $

:- module(tclp__utils,[ string_list_to_character_code_list/2,
                        instanciate_vars/1,
                        remove_spaces/2,
                        convert_list_to_atom/2,
                        convert_to_atom/2,
                        remove_all_constraints/1,
                        select_all/3,
                        true_member/2,
                        true_nth/3,
                        uniform_list/3,
                        unify_list/1,
			get_ordset/2,
			option/1,
			set_option/1,
			unset_option/1,
                        init_counter/1,
                        current_counter_value/2,
                        incr_counter/3,
                        print_constraints/1
                      ]).

:- use_module(library(lists)).
:- use_module(library(ordsets)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% characters and strings
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% converts a list of strings representing lines into a list of
%% characters codes
string_list_to_character_code_list('$empty',[]) :- !.
string_list_to_character_code_list([],[]):- !.
string_list_to_character_code_list(X,Codes) :-
        % in case an atom is given instead of a list
        atomic(X),!,atom_codes(X,Codes). 
string_list_to_character_code_list([ Line | RestOfLines ], Codes) :-
        append(Line, [ 10 | RestOfCodes ], Codes),
        string_list_to_character_code_list(RestOfLines, RestOfCodes).

%% delete the leading spaces
delete_leading([],[]).
delete_leading([ 32 | Rest ],Result) :- !,
        delete_leading(Rest,Result).
delete_leading(X,X).

%% removes the leading and traling spaces in codes
delete_leading_trailing(RawCodes,CleanCodes) :-
        delete_leading(RawCodes, CleanLead),
        reverse(CleanLead,RevCleanLead),
        delete_leading(RevCleanLead,RevClean),
        reverse(RevClean,CleanCodes).

%% removes the leading and traling spaces
remove_spaces(Raw, Clean) :-
        atom_codes(Raw,RawCodes),
        delete_leading_trailing(RawCodes,CleanCodes),
        atom_codes(Clean,CleanCodes).

%% convert atomic terms to their atom representations
convert_to_atom(X,_) :-
        var(X),!,throw(illegal_argument(convert_to_atom,1,variable)).
convert_to_atom(X,X) :- atom(X),!.
convert_to_atom(N,Atom) :-
        number(N),!,
        number_chars(N,Chars), atom_chars(Atom,Chars).

%% convert a list of atoms or numbers into a single atom
convert_list_to_atom(L,Atom) :-
        convert_list_to_atom(L,'',Atom).
convert_list_to_atom([],Atom,Atom).
convert_list_to_atom([X|L],Current,Result) :-
        convert_to_atom(X,Atom),
        atom_concat(Current,Atom,Next),
        convert_list_to_atom(L,Next,Result).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% variable printing
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% Computes the next codes
next_codes([C], [N]) :- !,
        N is C+1.
next_codes([ CurrentFirst | CurrentRest ], Result) :-
        next_codes(CurrentRest, [ NextSnd | RestNext ]),
        (
          NextSnd == 91 %% just after 'Z'
        ->
          NextFirst is CurrentFirst + 1,
          Result = [ NextFirst , 65 | RestNext ] %% 65 = 'A'
        ;
          Result = [ CurrentFirst, NextSnd | RestNext ]
        ).

convert_next([ 91 | Rest ],[ 65 , 65 | Rest ]) :- !.
convert_next(X,X).

%% Computes the next name for a variable
next_name(Current,Next) :-
        atom_codes(Current, CurrentCodes),
        next_codes(CurrentCodes, NextCodes1),
        convert_next(NextCodes1, NextCodes),
        atom_codes(Next,NextCodes).

%% instanciate variables to arbitrary, human readeable names
instanciate_vars(X) :-
        instanciate_vars(X,'A',_).

instanciate_vars(X,Avail,Next) :-
        var(X),!,
        X = Avail, next_name(Avail,Next).
instanciate_vars([],Avail,Avail) :- !.
instanciate_vars([ T | Rest], Avail, Next) :- !,
        instanciate_vars(T,Avail,Inter),
        instanciate_vars(Rest, Inter, Next).
instanciate_vars(T, Avail, Next) :-
        T =.. [ _ | Args ],
        instanciate_vars(Args, Avail, Next).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% CHR stuff
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% removes all the chr constraints of the given pattern
remove_all_constraints1([]).
remove_all_constraints1([#(_,Id)|Rest]) :-
        chr:remove_constraint(Id),
        remove_all_constraints1(Rest).
remove_all_constraints(Pattern) :-
        chr:findall_constraints(Pattern, Constraints),
        remove_all_constraints1(Constraints).

%% print the constraint corresponding to the given pattern
print_constraints(Pattern) :-
        chr:findall_constraints(Pattern, Constraints),
        print_constraints1(Constraints).
print_constraints1([]).
print_constraints1([Constraint#_|Rest]) :-
        write(Constraint),nl,
        print_constraints1(Rest).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Lists & ordered sets
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% remove all occurences of a term from a list
select_all(_,[],[]).
select_all(E,[E2|L],[E2|R]) :-
        E \= E2,!,
        select_all(E,L,R).
select_all(E,[_|L],R) :-
        select_all(E,L,R).

%% true if there is an element in the List which is == to X
true_member(_,[]) :-!,fail.
true_member(X,[Y|_]) :- X==Y,!.
true_member(X,[_|L]) :- true_member(X,L).

%% true_nth(N,L,E)
%% true if E is the Nth element of L. do not unify E with the element
true_nth(N,L,E) :- true_nth(1,L,E,N).
true_nth(_I,[],_E,_N) :- !,fail.
true_nth(I,[X|_Rest],E,I) :- X==E.
true_nth(I,[_X|Rest],E,N) :-
        I1 is I+1, true_nth(I1,Rest,E,N).

%% uniform_list(N,X,List).
%% List has length N and all its elements are unified with X
uniform_list(0,_,[]).
uniform_list(N,X,[X|Rest]) :-
        N1 is N-1, uniform_list(N1,X,Rest).

%% unify_list(L) unifies all elements of L one with another
unify_list([]).
unify_list([_]).
unify_list([A,A|L]) :- unify_list([A|L]).

%% unifies Ordset with the ordset corresponding to List,
%% that List itself if it is already an ordered set
get_ordset(List,List) :- is_ordset(List),!.
get_ordset(List,Ordset) :- list_to_ord_set(List,Ordset).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Options
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- dynamic option/1.

set_option(Option) :-
	assert(option(Option)),
        option_hook(Option).

unset_option(Option) :-
	retractall(option(Option)).

option_hook(disableterm) :-
        set_option(complete_order).
option_hook(_).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Counter
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- dynamic counter/2.

init_counter(Name) :-
        retractall(counter(Name,_)),
        assert(counter(Name,0)),!.

current_counter_value(Name,Value) :-
        counter(Name,Value).

incr_counter(Name,Value,NewValue) :-
        current_counter_value(Name,Value),
        retractall(counter(Name,_)),
        NewValue is Value+1,
        assert(counter(Name,NewValue)),!.

