%% tclp__variables_set_assoc.pl
%% Copyright (C) 2003 INRIA Emmanuel Coquery

%% This library is free software; you can redistribute it and/or
%% modify it under the terms of the GNU Lesser General Public
%% License as published by the Free Software Foundation; either
%% version 2.1 of the License, or (at your option) any later version.

%% This library is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
%% Lesser General Public License for more details.

%% You should have received a copy of the GNU Lesser General Public
%% License along with this library; if not, write to the Free Software
%% Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%                                                                           %%
%%                                   TCLP                                    %%
%%                                                                           %%
%%         Emmanuel Coquery, projet Contraintes, INRIA Rocquencourt          %%
%%                                                                           %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% $Header: /home/beaune/contraintes/coquery/cvs/contraintes/tclpchr/src/tclp/tclp__variables_set_assoc.pl,v 1.2 2003/07/25 15:27:13 coquery Exp $

:- module(tclp__variables_set_assoc,[associated/2, register/1]).

:- use_module(tclp(tclp__utils)).
%:- use_module(tclp__utils).
:- use_module(library(lists)).
:- use_module(library(chr)).
:- use_module(library(ordsets)).

%% unify Data with the data corresponding to Set in the tree, or adds Data
%% to the tree
tree_element([],Data,t(_,_,_,d(Data))) :- !.
tree_element([X|Rest],Data,t(V,Y,N,_)) :-
        var(Y),var(N),!,
        X=V, tree_element(Rest,Data,Y).
tree_element(Set,Data,t(V,Y,N,_)) :-
        true_member(V,Set) -> delete(Set,V,NewSet),tree_element(NewSet,Data,Y);
        tree_element(Set,Data,N).

%% lookup for Data using the key Set in the tree. fails if Data is not present
tree_lookup([],Data,t(_,_,_,Container)) :- !,
        nonvar(Container), Container = d(Data).
tree_lookup(_,_,t(_,Y,N,_)) :- var(Y),var(N),!,fail.
tree_lookup(Set,Data,t(V,Y,N,_)) :-
        true_member(V,Set) -> delete(Set,V,NewSet),tree_lookup(NewSet,Data,Y);
        tree_lookup(Set,Data,N).

%% CHR solver to handle the data structure.

handler tree_variable_sets.

constraints %% current tree
            current_tree/1,
            %% constraint put on any variable that appears on the set
            %% used to trigger recomputation of the tree if there is any
            %% unification
            set_variable/1.

%% if there is 2 constraints current_tree, unify the two trees
%% used to retreve the current_tree

current_tree(X) \ current_tree(Y) <=> X=Y.

%% check if some variables where declared
is_declared(X) :- chr:find_constraint(X,set_variable(X),_).
are_declared([]).
are_declared([X|L]) :- is_declared(X), are_declared(L).
register(X) :- is_declared(X),!.
register(X) :- set_variable(X).
declare_all([]).
declare_all([X|L]) :-  register(X), !, declare_all(L).

%% lookup or add the data corresponding to set or/and add data to the store
associated(Key,Data) :-
        (is_ordset(Key) -> Key=Set ; list_to_ord_set(Key,Set)),
        declare_all(Set),
        current_tree(Tree),
        tree_element(Set,Data,Tree).

%% unification of two registered variables
set_variable(X) \ set_variable(X) <=> trigger_unification(X).

%% reinitialize the set
reinit :-
        remove_all_constraints(set_variable(_)),
        remove_all_constraints(current_tree(_)).

trigger_unification(Duplicated) :- 
        current_tree(OldTree),
        build_new_tree(OldTree,[],NewTree,Duplicated,false),
        remove_all_constraints(current_tree(_)),
        current_tree(NewTree),!.

%% builds a new tree from the old one by prefix exploration.
%% The second argument represents a set which corresponds to
%% a subset in common to al the sets stored in the tree
build_new_tree(OldTree,_,_) :-
        var(OldTree), !. % case of a leaf
build_new_tree(t(V,Y,N,D), PartialSet, NewTree) :-
        (nonvar(D) -> % test if there is some data for the current set
            D=d(Data), tree_element(PartialSet, Data, NewTree) ;
            true),
        (true_member(V,PartialSet) ->
            YSet = PartialSet; % if the variable is in the set, don't add it
            YSet = [V|PartialSet]),
        build_new_tree(Y,YSet, NewTree),
        build_new_tree(N,PartialSet, NewTree).

%% builds a new tree from the old one by prefix exploration.
%% The second argument represents a set which corresponds to
%% a subset in common to al the sets stored in the tree
%% uses the last two elements to quickly find if a variable was already
%% in the set
build_new_tree(OldTree,_,_,_,_) :-
        var(OldTree), !. % case of a leaf
build_new_tree(t(V,Y,N,D), PartialSet, NewTree, Duplicated, Found) :-
        (nonvar(D) -> % test if there is some data for the current set
            D=d(Data), tree_element(PartialSet, Data, NewTree) ;
            true),
        ( Found
        ->
            YFound=true,
            ( Duplicated==V ->
                YSet = PartialSet;
                                % if the variable is in the set, don't add it
                YSet = [V|PartialSet])
        ;
            YSet = [V|PartialSet], 
            ( Duplicated==V -> YFound = true; YFound = false)
        ),
        build_new_tree(Y,YSet, NewTree, Duplicated, YFound),
        build_new_tree(N,PartialSet, NewTree, Duplicated, Found).


%% prints the current tree
print :- current_tree(T),
        nl,nl,write(T),nl,nl.
